/*
 * Copyright (C) 2001 Misha Nasledov <misha@nasledov.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <gtk/gtk.h>
#include <time.h>
#include <netinet/in.h>
#include "hotline.h"
#include "hxd.h"
#include "hx.h"
#include "hfs.h"
#include "network.h"
#include "rcv.h"
#include "chat.h"
#include "tasks.h"
#include "sound.h"
#include "files.h"

int nxfers = 0;
struct htxf_conn **xfers = 0;
void xfer_delete (struct htxf_conn *htxf);

struct file_update {
	guint32 type;
	guint32 pos;
	struct htxf_conn *htxf;
	char *buf;
	int bufsiz;
};

static void
xfer_pipe_close (int fd)
{
	close(fd);
	memset(&hxd_files[fd], 0, sizeof(struct hxd_file));
	hxd_fd_clr(fd, FDR);
	hxd_fd_del(fd);
}

static void
xfer_pipe_ready_read (int fd)
{
	ssize_t r;
	struct file_update fu;


	r = read(fd, &fu, sizeof(fu));
	if (r != sizeof(fu) || (r < 0 && errno != EINTR)) {
		play_sound(FILE_DONE);
		xfer_pipe_close(fd);
	} else {
		struct htxf_conn *htxf = fu.htxf;


		if (fu.type == 0) {
			/* update */
			htxf->total_pos = fu.pos;
			hx_output.file_update(&sessions[0], htxf);
		} else if (fu.type == 1) {
			/* done */
			hx_output.file_update(&sessions[0], htxf);
		} else {
			/* errors */
			int err = (int)fu.pos;
			switch (fu.type) {
			case 2:
				hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "get_thread: pipe: %s\n", strerror(err));
				break;
			case 3:
				hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "get_thread: fork: %s\n", strerror(err));
				break;
			case 4:
				hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "get_thread: dup2: %s\n", strerror(err));
				break;
			case 5:
				hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "%s: execve: %s\n", htxf->filter_argv[0], strerror(err));
				break;
			default:
				break;
			}
		}
	}
}

static int
resource_path (char *rsrcpath, const char *path, struct stat *statbuf)
{
	int i;
	size_t len;
	char pathbuf[MAXPATHLEN];
	struct stat sb;

	len = strlen(path);
	if (len + 16 >= MAXPATHLEN)
		return ENAMETOOLONG;
	strcpy(pathbuf, path);
	for (i = len - 1; i > 0; i--)
		if (pathbuf[i] == dir_char) {
			pathbuf[i++] = 0;
			break;
		}
	snprintf(rsrcpath, MAXPATHLEN, "%s/", i > 0 ? pathbuf : ".");
	if (stat(rsrcpath, &sb)) {
		if (statbuf)
			return ENOENT;
		if (mkdir(rsrcpath, S_IRUSR|S_IWUSR|S_IXUSR))
			return errno;
	}
	strcat(rsrcpath, "/");
	strcat(rsrcpath, &pathbuf[i]);
	strcat(rsrcpath, ".rsrc");
	if (statbuf && stat(rsrcpath, statbuf))
		return errno;

	return 0;
}

static int
resource_open (const char *path, int mode, int perm)
{
	int f = -1;

	if (1) {
		char rsrcpath[MAXPATHLEN];

		if (resource_path(rsrcpath, path, 0))
			return -1;
		f = open(rsrcpath, mode, perm);
	}

	return f;
}

size_t resource_len (const char *path)
{
	if (1) {
		char rsrcpath[MAXPATHLEN];
		struct stat sb;

		if (resource_path(rsrcpath, path, &sb))
			return 0;

		return sb.st_size;
	}

	return 0;
}

static int
finderinfo_path (char *infopath, const char *path, struct stat *statbuf)
{
	int i, len;
	char pathbuf[MAXPATHLEN];
	struct stat sb;

	len = strlen(path);
	if (len + 16 >= MAXPATHLEN)
		return ENAMETOOLONG;
	strcpy(pathbuf, path);
	for (i = len - 1; i > 0; i--)
		if (pathbuf[i] == dir_char) {
			pathbuf[i++] = 0;
			break;
		}
	snprintf(infopath, MAXPATHLEN, "%s/", i > 0 ? pathbuf : ".");
	if (stat(infopath, &sb)) {
		if (statbuf)
			return ENOENT;
		if (mkdir(infopath, S_IRUSR|S_IWUSR|S_IXUSR))
			return errno;
	}
	strcat(infopath, "/");
	strcat(infopath, &pathbuf[i]);
	strcat(infopath, ".fndrinfo");
	if (statbuf && stat(infopath, statbuf))
		return errno;

	return 0;
}

static int
hfsinfo_read (const char *path, struct hfs_cap_info *fi)
{
	char infopath[MAXPATHLEN];
	int f, r = 0;
	struct stat sb;

	if (!finderinfo_path(infopath, path, &sb)) {
		f = open(infopath, O_RDONLY);
		if (f >= 0) {
			r = read(f, fi, SIZEOF_HFS_CAP_INFO);
			close(f);
		}
	}

	return r;
}

static void
hfsinfo_write (const char *path, struct hfs_cap_info *fi)
{
	char infopath[MAXPATHLEN];
	int f;

	if (!finderinfo_path(infopath, path, 0)) {
		f = open(infopath, O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR);
		if (f >= 0) {
			fi->fi_magic1 = HFS_CAP_MAGIC1;
			fi->fi_version = HFS_CAP_VERSION;
			fi->fi_magic = HFS_CAP_MAGIC;
			fi->fi_datemagic = HFS_CAP_DMAGIC;
			fi->fi_datevalid = HFS_CAP_MDATE|HFS_CAP_CDATE;
			write(f, fi, SIZEOF_HFS_CAP_INFO);
			fsync(f);
			close(f);
		}
	}
}

inline int comment_len (const char *path)
{
	char infopath[MAXPATHLEN];
	int f, len = 0;
	struct stat sb;
	struct hfs_cap_info cap;

	if (!finderinfo_path(infopath, path, &sb)) {
		f = open(infopath, O_RDONLY);
		if (f >= 0) {
			if (read(f, &cap, SIZEOF_HFS_CAP_INFO) == SIZEOF_HFS_CAP_INFO)
				len = cap.fi_comln > 200 ? 200 : cap.fi_comln;
			close(f);
		}
	}

	return len;
}

static inline char *
suffix (const char *path)
{
	char *p = (char *)path + strlen(path);
	while (p-- > path)
		if (*p == '.')
			return p + 1;
	return 0;
}

static void
type_creator (guint8 *buf, const char *path)
{
	guint8 *tc;
	char *suff;
	char infopath[MAXPATHLEN];
	struct stat sb;

	if (!finderinfo_path(infopath, path, &sb)) {
		int r, fd = open(infopath, O_RDONLY);
		if (fd < 0)
			goto use_suffix;
		r = read(fd, buf, 8);
		close(fd);
		if (r != 8)
			goto use_suffix;
		return;
	}

use_suffix:
	tc = UNKNOWN_TYPECREA;
	if (!(suff = suffix(path)))
		goto dunno;


	if (!strcmp(suff, "jpg") || !strcmp(suff, "jpeg"))
		tc = "JPEGGKON";
	else if (!strcmp(suff, "png"))
		tc = "PNGfGKON";
	else if (!strcmp(suff, "gif"))
		tc = "GIFfGKON";
	else if (!strcmp(suff, "mp3") || !strcmp(suff, "mp2"))
		tc = "MP3 MAmp";
	else if (!strcmp(suff, "mpg") || !strcmp(suff, "mpeg"))
		tc = "MPEGTVOD";
	else if (!strcmp(suff, "mov"))
		tc = "MooVTVOD";
	else if (!strcmp(suff, "sit"))
		tc = "SITDSIT!";
	else if (!strcmp(suff, "zip"))
		tc = "ZIP ZIP ";
	else if (!strcmp(suff, "app") || !strcmp(suff, "sea"))
		tc = "APPLpeff";
	else if (!strcmp(suff, "img"))
		tc = "rohdWrap";
	else if (!strcmp(suff, "pict"))
		tc = "PICTGKON";

  dunno:
	memcpy(buf, tc, 8);
}

static void
ignore_signals (sigset_t *oldset)
{
	sigset_t set;


	sigfillset(&set);
	sigprocmask(SIG_BLOCK, &set, oldset);
}

static void
unignore_signals (sigset_t *oldset)
{
	sigprocmask(SIG_SETMASK, oldset, 0);
}

void xfer_go (struct htxf_conn *htxf)
{
	char *rfile;
	guint16 hldirlen;
	guint8 *hldir;
	guint32 data_size = 0, rsrc_size = 0;
	guint8 rflt[74];
	struct stat sb;

	if (htxf->gone)
		return;

	htxf->gone = 1;

	if(htxf->type == XFER_GET) {
/*		hx_htlc.nr_gets++; */
	}
	else if(htxf->type == XFER_PUT) {
/*		hx_htlc.nr_puts++; */
	}
	if (htxf->type == XFER_GET) {
		if (!stat(htxf->path, &sb))
			data_size = sb.st_size;
		rsrc_size = resource_len(htxf->path);

		rfile = dirchar_basename(htxf->remotepath);
		if (data_size || rsrc_size) {
			memcpy(rflt, "\
                          RFLT\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
                          \0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\2\
                          DATA\0\0\0\0\0\0\0\0\0\0\0\0\
                          MACR\0\0\0\0\0\0\0\0\0\0\0\0", 74);
			S32HTON(data_size, &rflt[46]);
			S32HTON(rsrc_size, &rflt[62]);
			htxf->data_pos = data_size;
			htxf->rsrc_pos = rsrc_size;
		}
		task_new(&sessions[0].htlc, rcv_task_file_get, htxf, 0, "xfer_go");
		if (rfile != htxf->remotepath) {
			hldir = path_to_hldir(htxf->remotepath, &hldirlen, 1);
			hlwrite(&sessions[0].htlc, HTLC_HDR_FILE_GET, 0, (data_size || rsrc_size) ? 3 : 2,
					HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
					HTLC_DATA_DIR, hldirlen, hldir,
					HTLC_DATA_RFLT, 74, rflt);
			g_free(hldir);
		} else {
			hlwrite(&sessions[0].htlc, HTLC_HDR_FILE_GET, 0, (data_size || rsrc_size) ? 2 : 1,
					HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
					HTLC_DATA_RFLT, 74, rflt);
		}
	} else {
		rfile = basename(htxf->path);
		hldir = path_to_hldir(htxf->remotepath, &hldirlen, 1);
		if (exists_remote(htxf->remotepath)) {
			task_new(&sessions[0].htlc, rcv_task_file_put, htxf, 0, "xfer_go");
			hlwrite(&sessions[0].htlc, HTLC_HDR_FILE_PUT, 0, 3,
					HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
					HTLC_DATA_DIR, hldirlen, hldir,
					HTLC_DATA_FILE_PREVIEW, 2, "\0\1");
		} else {
			task_new(&sessions[0].htlc, rcv_task_file_put, htxf, 0, "xfer_go");
			hlwrite(&sessions[0].htlc, HTLC_HDR_FILE_PUT, 0, 2,
					HTLC_DATA_FILE_NAME, strlen(rfile), rfile,
					HTLC_DATA_DIR, hldirlen, hldir);
		}
		g_free(hldir);
	}
}

int xfer_go_timer (void *__arg)
{


	xfer_go((struct htxf_conn *)__arg);



	return 0;
}

struct htxf_conn *xfer_new (const char *path, const char *remotepath, guint16 type)
{
	struct htxf_conn *htxf;

	htxf = g_malloc0(sizeof(struct htxf_conn));
	strcpy(htxf->remotepath, remotepath);
	strcpy(htxf->path, path);
	htxf->type = type;
	htxf->queue = -1;

	xfers = g_realloc(xfers, (nxfers + 1) * sizeof(struct htxf_conn *));
	xfers[nxfers] = htxf;
	nxfers++;


	htxf->htlc = &sessions[0].htlc;
	htxf->total_pos = 0;
	htxf->total_size = 1;
	hx_output.file_update(&sessions[0], htxf);



	if(nxfers == 1 || !gtkhx_prefs.queuedl) {
		xfer_go(htxf);
	}




	return htxf;
}

void xfer_up(int num)
{
	struct htxf_conn *tmp;

	tmp = xfers[num-1];
	xfers[num-1] = xfers[num];
	xfers[num] = tmp;
}

int xfer_down(int num)
{
	struct htxf_conn *tmp;


	if(nxfers-1 == num) {
		return 1;
	}


	tmp = xfers[num+1];
	xfers[num+1] = xfers[num];
	xfers[num] = tmp;

	return 0;
}

int xfer_num (struct htxf_conn *htxf)
{
	int i;


	for(i = 0; i < nxfers; i++) {
		if(xfers[i] == htxf) {
			return i;
		}
	}

	return -1;
}

static void xfer_pipe_error(struct htxf_conn *htxf, guint32 type, int err)
{
	struct file_update fu;


	fu.type = type;
	fu.pos = err;
	fu.htxf = htxf;
	write(htxf->pipe, &fu, sizeof(fu));
}

static void
xfer_pipe_done (struct htxf_conn *htxf)
{
    struct file_update fu;

    fu.type = 1;
    fu.pos = htxf->total_pos;
    fu.htxf = htxf;
    write(htxf->pipe, &fu, sizeof(fu));
}

static void xfer_pipe_update(struct htxf_conn *htxf)
{
	struct file_update fu;

	fu.type = 0;
	fu.pos = htxf->total_pos;
	fu.htxf = htxf;
	write(htxf->pipe, &fu, sizeof(fu));
}

static int
rd_wr (int rd_fd, int wr_fd, guint32 data_len, struct htxf_conn *htxf)
{
	int r, pos, len;
	guint8 *buf;
	size_t bufsiz;


	bufsiz = 0xf000;
	buf = g_malloc(bufsiz);
	if (!buf)
		return 111;
	while (data_len) {
		if ((len = read(rd_fd, buf, (bufsiz < data_len) ? bufsiz : data_len)) < 1)
			return len ? errno : EIO;
		pos = 0;
		while (len) {
			if ((r = write(wr_fd, &(buf[pos]), len)) < 1)
				return errno;
			pos += r;
			len -= r;
			htxf->total_pos += r;
			xfer_pipe_update(htxf);
		}
		data_len -= pos;
	}
	g_free(buf);

	return 0;
}

typedef int thread_return_type;


static thread_return_type
get_thread (void *__arg)
{
	struct htxf_conn *htxf = (struct htxf_conn *)__arg;
	guint32 pos, len, tot_len;
	int s, f, r, retval = 0;
	guint8 typecrea[8], buf[1024];
	struct hfs_cap_info fi;

	s = htxf_connect(htxf);
	if (s < 0) {
		retval = s;
		goto ret;
	}

	len = 40;
	pos = 0;
	while (len) {
		if ((r = read(s, &(buf[pos]), len)) < 1) {
			retval = errno;
			goto ret;
		}
		pos += r;
		len -= r;
		htxf->total_pos += r;
	}
	pos = 0;
	len = (buf[38] ? 0x100 : 0) + buf[39];
	len += 16;
	tot_len = 40 + len;
	while (len) {
		if ((r = read(s, &(buf[pos]), len)) < 1) {
			retval = errno;
			goto ret;
		}
		pos += r;
		len -= r;
		htxf->total_pos += r;
		xfer_pipe_update(htxf);
	}
	memcpy(typecrea, &buf[4], 8);
	memset(&fi, 0, sizeof(struct hfs_cap_info));
	fi.fi_comln = buf[73 + buf[71]];
	memcpy(fi.fi_fndr, "HTftHTLC", 8);
	memcpy(fi.fi_comnt, &buf[74 + buf[71]], fi.fi_comln);
	*((guint32 *)(&buf[56])) = hfs_m_to_htime(*((guint32 *)(&buf[56])));
	*((guint32 *)(&buf[64])) = hfs_m_to_htime(*((guint32 *)(&buf[64])));
	memcpy(&fi.fi_ctime, &buf[56], 4);
	memcpy(&fi.fi_mtime, &buf[64], 4);
	hfsinfo_write(htxf->path, &fi);

	L32NTOH(len, &buf[pos - 4]);
	tot_len += len;
	if ((f = open(htxf->path, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR)) < 0) {
		retval = errno;
		goto ret;
	}
	if (!len)
		goto get_rsrc;
	if (htxf->data_pos)
		lseek(f, htxf->data_pos, SEEK_SET);
	if (htxf->filter_argv) {
		pid_t pid;
		int pfds[2];

		if (pipe(pfds)) {
			xfer_pipe_error(htxf, 2, errno);
			goto no_filter;
		}
		pid = fork();
		switch (pid) {
			case -1:
				xfer_pipe_error(htxf, 3, errno);
				close(pfds[0]);
				close(pfds[1]);
				goto no_filter;
			case 0:
				close(pfds[1]);
				if (dup2(pfds[0], 0) == -1 || dup2(f, 1) == -1) {
					xfer_pipe_error(htxf, 4, errno);
					_exit(1);
				}
				close(2);
				if (htxf->data_pos)
					lseek(0, htxf->data_pos, SEEK_SET);
				execve(htxf->filter_argv[0], htxf->filter_argv, hxd_environ);
				xfer_pipe_error(htxf, 5, errno);
				_exit(127);
			default:
				close(pfds[0]);
				close(f);
				f = pfds[1];
		}
	}
no_filter:
	retval = rd_wr(s, f, len, htxf);
	fsync(f);
	close(f);
	if (retval)
		goto ret;
get_rsrc:
	if (tot_len >= htxf->total_size)
		goto done;
	pos = 0;
	len = 16;
	while (len) {
		if ((r = read(s, &(buf[pos]), len)) < 1) {
			retval = errno;
			goto ret;
		}
		pos += r;
		len -= r;
		htxf->total_pos += r;
		xfer_pipe_update(htxf);
	}
	L32NTOH(len, &buf[12]);
	if (!len)
		goto done;
	if ((f = resource_open(htxf->path, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR)) < 0) {
		retval = errno;
		goto ret;
	}
	if (htxf->rsrc_pos)
		lseek(f, htxf->rsrc_pos, SEEK_SET);
	retval = rd_wr(s, f, len, htxf);
	if (retval)
		goto ret;
	close(f);
done:

	memcpy(fi.fi_fndr, typecrea, 8);
	hfsinfo_write(htxf->path, &fi);
	xfer_pipe_done(htxf);

ret:
	close(s);

	return (thread_return_type)retval;
}

static thread_return_type
put_thread (void *__arg)
{
	struct htxf_conn *htxf = (struct htxf_conn *)__arg;
	int s, f, retval = 0;
	guint8 buf[512];
	struct hfs_cap_info fi;

	s = htxf_connect(htxf);
	if (s < 0) {
		retval = s;
		goto ret;
	}

	memcpy(buf, "\
FILP\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\2INFO\0\0\0\0\0\0\0\0\0\0\0^AMAC\
TYPECREA\
\0\0\0\0\0\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\
\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\
\7\160\0\0\0\0\0\0\7\160\0\0\0\0\0\0\0\0\0\3hxd", 115);
	hfsinfo_read(htxf->path, &fi);
	if (htxf->rsrc_size - htxf->rsrc_pos)
		buf[23] = 3;
	if (65 + fi.fi_comln + 12 > 0xff)
		buf[38] = 1;
	buf[39] = 65 + fi.fi_comln + 12;
	type_creator(&buf[44], htxf->path);
	*((guint32 *)(&buf[96])) = hfs_h_to_mtime(*((guint32 *)(&fi.fi_ctime)));
	*((guint32 *)(&buf[104])) = hfs_h_to_mtime(*((guint32 *)(fi.fi_mtime)));
	buf[116] = fi.fi_comln;
	memcpy(buf+117, fi.fi_comnt, fi.fi_comln);
	memcpy(&buf[117] + fi.fi_comln, "DATA\0\0\0\0\0\0\0\0", 12);
	S32HTON((htxf->data_size - htxf->data_pos), &buf[129 + fi.fi_comln]);
	if (write(s, buf, 133 + fi.fi_comln) != 133 + fi.fi_comln) {
		retval = errno;
		goto ret;
	}
	htxf->total_pos += 133 + fi.fi_comln;
	if (!(htxf->data_size - htxf->data_pos))
		goto put_rsrc;
	if ((f = open(htxf->path, O_RDONLY)) < 0) {
		retval = errno;
		goto ret;
	}
	if (htxf->data_pos)
		lseek(f, htxf->data_pos, SEEK_SET);
	retval = rd_wr(f, s, htxf->data_size, htxf);
	if (retval)
		goto ret;
	close(f);

put_rsrc:
	if (!(htxf->rsrc_size - htxf->rsrc_pos))
		goto done;
	memcpy(buf, "MACR\0\0\0\0\0\0\0\0", 12);
	S32HTON(htxf->rsrc_size, &buf[12]);
	if (write(s, buf, 16) != 16) {
		retval = 0;
		goto ret;
	}
	htxf->total_pos += 16;
	if ((f = resource_open(htxf->path, O_RDONLY, 0)) < 0) {
		retval = errno;
		goto ret;
	}
	if (htxf->rsrc_pos)
		lseek(f, htxf->rsrc_pos, SEEK_SET);
	retval = rd_wr(f, s, htxf->rsrc_size, htxf);
	if (retval)
		goto ret;
	close(f);

done:

	xfer_pipe_done(htxf);

ret:
	close(s);

	return (thread_return_type)retval;
}

void xfer_ready_write (struct htxf_conn *htxf)
{
	sigset_t oldset;
	struct sigaction act, tstpact, contact;
	pid_t pid;
	int pfds[2];

	if (pipe(pfds)) {
		hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "xfer: pipe: %s\n", strerror(errno));
		goto err_fd;
	}
	if (pfds[0] >= hxd_open_max) {
		hx_printf_prefix(&sessions[0].htlc, 0, INFOPREFIX, "%s:%d: %d >= hxd_open_max (%d)\n", __FILE__, __LINE__, pfds[0], hxd_open_max);
		goto err_pfds;
	}
	htxf->pipe = pfds[1];

	ignore_signals(&oldset);
	act.sa_flags = 0;
	act.sa_handler = SIG_DFL;
	sigfillset(&act.sa_mask);
	sigaction(SIGTSTP, &act, &tstpact);
	sigaction(SIGCONT, &act, &contact);
	pid = fork();
	if(pid == 0) {
		_exit(htxf->type == XFER_GET ? get_thread(htxf) : put_thread(htxf));
	}


	sigaction(SIGTSTP, &tstpact, 0);
	sigaction(SIGCONT, &contact, 0);
	unignore_signals(&oldset);


	if(pid == -1) {
		goto err_pfds;
	}
	htxf->pid = pid;

	close(pfds[1]);
	hxd_files[pfds[0]].ready_read = xfer_pipe_ready_read;
	hxd_files[pfds[0]].conn.htxf = htxf;
	hxd_fd_add(pfds[0]);
	hxd_fd_set(pfds[0], FDR);

	return;


err_pfds:
	close(pfds[0]);
	close(pfds[1]);

err_fd:

	xfer_delete(htxf);
}

void xfer_tasks_update (struct htlc_conn *htlc)
{
	int i;



	for (i = 0; i < nxfers; i++) {
		if (xfers[i]->htlc == htlc)
			hx_output.file_update(&sessions[0], xfers[i]);
	}


}

void xfers_delete_all (void)
{
	struct htxf_conn *htxf;
	int i;

	for (i = 0; i < nxfers; i++) {
		htxf = xfers[i];
		if (htxf->pid) {
			kill(htxf->pid, SIGKILL);
		} else {
			timer_delete_ptr(htxf);
			g_free(htxf);
		}
	}
	nxfers = 0;
}

void xfer_delete (struct htxf_conn *htxf)
{
	int i;

	for (i = 0; i < nxfers; i++) {
		if (xfers[i] == htxf) {
			if(htxf->pid) {
				kill(htxf->pid, SIGKILL);
			}
			else {
				timer_delete_ptr(htxf);
				g_free(htxf);
				if (nxfers > (i+1))
					memcpy(xfers+i, xfers+i+1, (nxfers-(i+1)) * sizeof(struct htxf_conn *));
				nxfers--;
				if (nxfers) {
					xfer_go(xfers[0]);
				}
			}
#ifdef USE_IPV6
			freeaddrinfo(htxf->listen_addr);
#endif
			break;
		}
	}
}

static int
free_htxf (void *__arg)
{
	int i;
	pid_t pid = (pid_t)__arg;
	struct htxf_conn *htxfp;



	for (i = 0; i < nxfers; i++) {
		htxfp = xfers[i];
		if (htxfp->pid == pid) {
			htxfp->pid = 0;
			xfer_delete(htxfp);
			break;
		}
	}



	return 0;
}


struct htxf_conn *htxf_with_ref(guint16 ref)
{
	int i;

	for(i = 0; i < nxfers; i++) {
		if(xfers[i]->ref == ref) {
			return xfers[i];
		}
	}

	return 0;
}

void
hlclient_reap_pid (pid_t pid, int status)
{
	struct htxf_conn *htxfp;
	int i;

	for (i = 0; i < nxfers; i++) {
		htxfp = xfers[i];
		if (htxfp->pid == pid) {
			timer_add_secs(0, free_htxf, (void *)pid);
			return;
		}
	}
}
